# RROS设计文档

## 1.概述

### 1.1 RROS 介绍

RROS 是由 Rust 编写的一款基于 RISC-V 架构的多核同步宏内核模块化操作系统, 整体设计基于 rCore-Tutorial-v3。
该操作系统支持 FAT32 文件系统, 通过模块化的代码组织方式, 将宏内核的各个部分独立为不同的模块, 明确模块间的接口规范, 
有效降低了功能模块之间的耦合度, 从而显著简化了开发与调试的流程。这种设计理念不仅提升了开发人员的工作效率, 也为后续系统功能的扩展与优化提供了灵活性和可维护性。

RROS 的设计目标不仅在于功能的完善, 还注重性能的优化与实现高质量的代码。
功能方面, RROS 为运行在其上的用户态程序提供了完备且可靠的系统调用支持, 并借此通过了初赛阶段的所有系统调用测试用例, 充分验证了其稳定性和功能正确性；
性能优化方面, RROS在文件系统层面通过引入块缓存技术, 有效加速了文件读写操作, 显著减少了对磁盘 I/O 的访问次数, 从而提高了整体系统的运行效率。 

### 1.2 RROS整体架构
```
代码框架:
├── bootloader                       # 引导模块
│   ├── opensbi-qemu                 # OpenSBI的可执行文件
│   └── opensbi-qemu.bin             # OpenSBI的二进制文件
├── img                              
│   └── sdcard.img                   # 内核镜像文件
├── os                               # 操作系统内核
│   ├── Cargo.lock
│   ├── Cargo.toml
│   ├── Makefile
│   ├── build
│   ├── build.rs                     # 动态加载用户程序
│   ├── build_c.rs
│   ├── matrix.S
│   ├── src
│   │   ├── boards                   # qemu 相关设置
│   │   │   ├── mod.rs
│   │   │   └── qemu.rs
│   │   ├── config.rs                # 默认参数设置
│   │   ├── console.rs               
│   │   ├── drivers                  # 块设备驱动
│   │   │   ├── block
│   │   │   │   ├── block_dev.rs
│   │   │   │   ├── mod.rs
│   │   │   │   └── virtio_blk.rs
│   │   │   └── mod.rs
│   │   ├── entry.S                  # 启动程序
│   │   ├── fat32                    # fat32文件系统
│   │   │   ├── block_cache.rs
│   │   │   ├── dentry.rs
│   │   │   ├── fat.rs
│   │   │   ├── file.rs
│   │   │   ├── fs.rs
│   │   │   ├── inode.rs
│   │   │   ├── layout.rs
│   │   │   ├── mod.rs
│   │   │   └── time.rs
│   │   ├── fs                       # 文件相关数据结构设置
│   │   │   ├── inode.rs
│   │   │   ├── mod.rs
│   │   │   ├── os_inode.rs
│   │   │   ├── path.rs
│   │   │   ├── pipe.rs
│   │   │   └── stdio.rs
│   │   ├── index_list               # 索引列表相关实现
│   │   │   ├── listdrainiter.rs
│   │   │   ├── listends.rs
│   │   │   ├── listindex.rs
│   │   │   ├── listiter.rs
│   │   │   ├── listnode.rs
│   │   │   └── mod.rs
│   │   ├── lang_items.rs
│   │   ├── link_app.S               # 加载用户程序
│   │   ├── linker.ld                # 链接脚本
│   │   ├── loader.rs                # 加载用户程序
│   │   ├── logging.rs               # 日志系统
│   │   ├── main.rs                  # 主函数
│   │   ├── mm                       # 内存管理模块
│   │   │   ├── address.rs
│   │   │   ├── frame_allocator.rs
│   │   │   ├── heap_allocator.rs
│   │   │   ├── memory_set.rs
│   │   │   ├── mod.rs
│   │   │   └── page_table.rs
│   │   ├── mutex                    # 互斥锁模块
│   │   │   ├── mod.rs
│   │   │   └── spin_mutex.rs
│   │   ├── sbi.rs                   # sbi调用
│   │   ├── sched                    # 调度模块
│   │   │   ├── cfs.rs
│   │   │   ├── fifo.rs
│   │   │   ├── mod.rs
│   │   │   └── prio.rs
│   │   ├── syscall                  # 系统调用模块
│   │   │   ├── fs.rs
│   │   │   └── mod.rs
│   │   ├── task                     # 任务模块
│   │   │   ├── aux.rs
│   │   │   ├── context.rs
│   │   │   ├── id.rs                # 定义进程号相关信息
│   │   │   ├── kstack.rs            # 定义进程相关栈信息
│   │   │   ├── mod.rs               # 定义进程相关功能与结构
│   │   │   ├── processor.rs
│   │   │   ├── scheduler.rs         # 调度器
│   │   │   ├── switch.S
│   │   │   ├── switch.S.bak
│   │   │   └── switch.rs            # 进程切换工具
│   │   ├── timer.rs                 # 定时器设置
│   │   ├── trap                     # 陷阱模块
│   │   │   ├── context.rs
│   │   │   ├── irq.rs
│   │   │   ├── mod.rs
│   │   │   └── trap.S
│   │   └── utils.rs                 # 工具函数
├── user
│   ├── Cargo.lock
│   ├── Cargo.toml
│   ├── Makefile
│   ├── src
│   │   ├── bin                      
│   │   │   ├── exit.rs
│   │   │   ├── fantastic_text.rs
│   │   │   ├── forkexec.rs
│   │   │   ├── forktest.rs
│   │   │   ├── forktest2.rs
│   │   │   ├── forktest_simple.rs
│   │   │   ├── forktree.rs
│   │   │   ├── hello_world.rs
│   │   │   ├── initproc.rs
│   │   │   ├── matrix.rs
│   │   │   ├── my_hello_world.rs
│   │   │   ├── sleep.rs
│   │   │   ├── sleep_simple.rs
│   │   │   ├── stack_overflow.rs
│   │   │   ├── user_shell.rs
│   │   │   ├── usertests-simple.rs
│   │   │   ├── usertests.rs
│   │   │   └── yield.rs
│   │   ├── console.rs
│   │   ├── lang_items.rs
│   │   ├── lib.rs
│   │   ├── linker.ld
│   │   └── syscall.rs
```
OS逻辑框架: ![alx_text](./img/OS整体框架.png)

## 2.进程管理

### 2.1 进程相关数据结构

在 RROS 操作系统中, 进程是资源分配和 CPU 调度的基本单位, 承担着系统运行的核心功能。每个进程都具有独立的地址空间, 用于隔离运行环境, 从而确保进程之间的操作互不干扰。
此外, 每个进程在其独立的地址空间中分配有专属的栈空间和一组与之关联的资源, 如内存区域、文件描述符等, 用于存储局部变量、函数调用的返回地址以及中断处理相关的数据。
这些资源的分配和管理在操作系统内核中通过精心设计的机制进行协调, 以确保进程能够高效、可靠地完成任务。

### 2.1.1 任务结构
```
pub struct Task {
    pub kstack: KernelStack,               //任务栈指针
    pub tid: usize,                        //任务号
    pub inner: SpinNoIrqLock<TaskInner>,   //任务的内部状态和元数据
}
```
Task 结构体各个字段的含义如下：
| 字段名 |                             含义                              |
| :----- | :-----------------------------------------------------------: |
| kstack | KernelStack结构体, 其中包含了由操作系统为其分配的栈的栈顶指针 |
| tid    |               表示该任务的唯一标识符(任务 ID)               |
| inner  | 用于存储任务的内部状态和元数据信息, 由自旋锁保证其并发安全性  |

> SpinNoIrqLock 是一种自旋锁, 适用于短时间的锁定操作。由于在持有锁时发生中断, 可能导致死锁, 
而 SpinNoIrqLock 在将数据上锁的同时禁止中断, 因此能够保证在并发的情况下进程的内部状态与元数据不会因多核与中断环境下的偶发因素而遭到修改与破坏, 保证了数据的统一性

```
pub struct TaskInner {
    pub memory_set: MemorySet,             //内存相关信息
    pub task_status: TaskStatus,           //任务状态
    pub parent: Option<Weak<Task>>,        //父进程指针
    pub children: Vec<Arc<Task>>,          //子进程序列
    pub exit_code: i32,                    //退出码
    pub fd_table: Vec<Option<Arc<dyn File + Send + Sync>>>,  //文件打开表
    pub cwd: Path,                         //当前进程路径
}

```
TaskInner 结构体各个字段的含义如下：
| 字段名      |                            含义                             |
| :---------- | :---------------------------------------------------------: |
| memory_set  |     MemorySet结构体, 其中包含了进程所占页表与索引项信息     |
| task_status | TaskStatus枚举体, 其中包含了进程可能的状态(详情参见2.1.2) |
| parent      |     指向当前进程父进程的弱引用, initproc的parent为None      |
| children    |                存储其子进程引用指针的序列,                  |
| exit_code   |                        进程的退出码                         |
| fd_table    |                      进程的文件打开表                       |
| cwd         |            Path结构体, 其中包含了进程的当前路径             |

> 在 Rust 中, Arc(原子引用计数)是一种智能指针, 允许多个所有者安全地共享同一个数据, 同时维护该数据的引用计数。当最后一个引用被释放时, 数据才会真正被回收。  
> 相比之下, Weak(弱引用)是一种特殊的引用形式, 它不会增加对象的引用计数, 因此不会影响对象的生命周期。弱引用的核心作用是避免循环引用导致的内存泄漏问题。
通过使用 Weak 引用, 可以在某些情况下访问共享数据, 而无需干扰其生命周期管理。
  
在 RROS 操作系统的进程管理中, 父子进程关系的设计巧妙地运用了 Arc 和 Weak 的特性。
在父进程的子指针中使用 Arc 引用, 这确保了子进程在其生命周期内不会被意外销毁, 保证了父进程能够可靠地追踪和管理子进程。而在父进程的引用中, 子进程的父指针采用 Weak 引用。
这种设计有效避免了循环引用的发生, 因为 Weak 不会增加父进程的引用计数, 因此即使子进程持有父指针, 也不会阻止父进程的正常释放。

### 2.1.2 任务上下文信息

在RROS的设计与实现中, 任务的上下文环境管理是一个关键环节, 它涉及到任务状态的保存与切换。在这个过程中, 与当前任务上下文环境直接相关的结构体主要有两个:
TrapContext 和 TaskContext。它们在功能和用途上有所区别, 分别负责不同的上下文管理任务。  
TrapContext 结构体用于保存和管理中断处理过程中的任务上下文。当发生中断、异常或系统调用时, 操作系统需要保存当前任务的状态, 以便在中断处理完成后恢复任务的正常执行。  

```
pub struct TrapContext {
    pub x: [usize; 32],                    // 通用寄存器信息
    pub sstatus: Sstatus,                  // sstatus寄存器状态
    pub sepc: usize,                       // pc信息
}
```
TrapContext 结构体各个字段的含义如下：
| 字段名  |                            含义                             |
| :------ | :---------------------------------------------------------: |
| x       | 通用寄存器信息, 其中x[1]为ra, x[2]为sp, x[4]为tp, x[10]为a0 |
| sstatus |         Sstatus枚举体, 其中存储了sstatus中的位信息          |
| sepc    |          sepc寄存器信息, 用于记录发生中断时的PC值           |

TaskContext 结构体用于保存和管理任务切换过程中的上下文。当任务从运行状态被切换出去时, 其运行状态需要被保存；
当任务再次被调度时, 需要从保存的状态恢复。TaskContext 是任务调度机制的核心数据结构。

```
pub struct TaskContext {
    ra: usize,                             // 返回地址
    tp: usize,                             // 进程指针
    s: [usize; 12],                        // 被调用者保存寄存器
    satp: usize,                           // 页表指针
}
```
TaskContext 结构体各个字段的含义如下：
| 字段名 |                                                         含义                                                          |
| :----- | :-------------------------------------------------------------------------------------------------------------------: |
| ra     | ra 保存了任务切换时的返回地址。在任务切换的汇编代码中, 用于记录切出的任务执行到的位置, 切回该任务时将从此处继续执行。 |
| tp     |                                            tp 通常指向当前任务的任务结构体                                            |
| s      |                                           s 保存了任务的 被调用者保存寄存器                                           |
| satp   |                                         satp 寄存器保存了当前任务的页表信息。                                         |


### 2.1.3 任务状态
```
pub enum TaskStatus {
    Ready,
    Running,
    Zombie,
}
```

在该操作系统的任务管理中, 任务的生命周期被划分为三种主要的状态：就绪态(Ready)、运行态(Running)和终止态(Zombie)。
这些状态之间的转换与任务执行的不同阶段息息相关, 构成了进程管理的核心逻辑。

| 任务状态 |                              状态描述                              |
| :------- | :----------------------------------------------------------------: |
| Ready    | 表示进程已经获得了除 CPU 外的所有资源, 随时可以被调度到 CPU 上执行 |
| Running  |  表示进程获得了CPU的使用权, 此状态下, 任务占用 CPU 并执行其代码。  |
| Zombie   |            表示进程已经完成了任务或被操作系统强制终止。            |

### 2.2 进程创建
### 2.2.1 初始进程initproc创建
```
lazy_static! {
    /// Global process that init user shell
    pub static ref INITPROC: Arc<Task> = Task::new(get_app_data_by_name("initproc").unwrap());
}
```
```
pub fn add_initproc() {
    add_task(INITPROC.clone());
    // 设置tp寄存器指向INITPROC
    let initproc_tp = Arc::as_ptr(&INITPROC) as usize;
    unsafe {
        asm!("mv tp, {}", in(reg) initproc_tp);
    }
}
```
在 Rust 编程语言中, lazy_static 是一个流行的工具, 用于实现惰性静态初始化。它允许用户定义具有 'static 生命周期的全局变量, 
但这些变量的初始化会被延迟到它们第一次被访问时, 而不是在程序启动时立即初始化。这种特性不仅可以避免初始化顺序问题, 还能在某些情况下提高程序性能, 
尤其是在初始化过程较为复杂或代价较高的场景中。在这段代码中, 初始进程 INITPROC 就是通过 lazy_static 宏创建的。
INITPROC 是系统中第一个被定义的任务, 它的作用通常是加载 Shell 或执行其他初始化任务。使用 lazy_static 创建 INITPROC 的好处是, 初始化操作仅在真正需要时进行, 避免了在程序启动时耗费不必要的资源。  

在操作系统的主函数中, 通过调用 add_initproc 方法, 将初始进程 INITPROC 进行初始化, 并装载到任务调度系统的就绪队列中。
INITPROC 是系统的第一个任务, 其初始化过程包括将其关联到任务队列并设置关键资源。这一操作确保 INITPROC 被任务调度器管理, 任务调度器在运行时会从就绪队列中选择任务执行, 而 INITPROC 是被优先调度的第一个任务。当 INITPROC 被分配到 CPU 上运行时, 它的主要职责是加载用户 Shell, 提供一个交互式的命令行环境。这一过程是操作系统从内核启动到用户态运行的重要过渡阶段。

### 2.2.2 其余进程创建

fork 是一个系统调用, 用于创建一个新进程。它通过复制当前进程的整个内存空间、文件描述符和执行上下文, 生成一个与父进程几乎完全相同的子进程。
fork 是操作系统实现多任务的基础机制之一,在 RROS 操作系统中, 新的进程同样通过 fork 函数创建。其主要执行流程如下：

1. 复制内存空间  
fork 首先为子进程复制父进程的内存空间, 包括地址空间中的所有有效页面。这一步使用了写时复制(Copy-On-Write, COW)机制, 即仅在需要写入时才真正复制内存页面。通过这一优化, 系统显著减少了不必要的内存分配开销, 提高了资源利用率。

2. 复制内核栈中的 TrapContext  
TrapContext 是保存进程陷入内核模式时上下文的关键数据结构。fork 会为子进程复制父进程的 TrapContext, 确保子进程能够在相同的用户态上下文中运行。然而, 为了避免冗余, 父进程在内核模式下的函数调用栈帧不会被复制, 这使得子进程在执行时能够以一个清晰的初始状态运行。

3. 复制文件描述符表和工作目录  
子进程会继承父进程的文件描述符表(fd_table)以及当前工作目录(cwd)。这确保了子进程能够共享父进程打开的文件或资源, 从而实现父子进程间的资源共享。

4. 创建子进程的 Task 并分配新的线程 ID(tid)  
操作系统为子进程创建一个新的任务控制块(Task), 这是管理进程生命周期的核心数据结构。同时, 系统为子进程分配一个唯一的线程 ID(tid), 以便区分不同的进程并支持调度管理。

5. 初始化子进程的 TrapContext  
在完成 TrapContext 的复制后, 系统对其进行初始化和适配, 使得子进程在用户态恢复运行时能够正确执行。从子进程视角看, fork 调用会返回 0, 以便区分其与父进程的执行逻辑。

6. 设置子进程的 TaskContext  
操作系统为子进程分配和初始化新的任务上下文(TaskContext)。这是子进程在内核态的运行入口, 通常用于设置从 fork 返回到用户态的路径。 

```
pub fn fork(self: &Arc<Self>) -> Arc<Self> {
        let mut parent_inner = self.inner.lock();
        // 复制内存空间
        let memory_set = MemorySet::from_existed_user(&parent_inner.memory_set);
        // 只复制内核栈中的`TrapContext`, 不复制父进程在内核中函数调用的栈帧
        let dst_kstack_top = kstack_alloc();
        log::info!("self.kstack.0: {:#x}", self.kstack.0);
        let src_kstack_top = get_stack_top_by_sp(self.kstack.0);
        log::info!(
            "[Task::fork] src_kstack_top: {:#x}, dst_kstack_top: {:#x}",
            src_kstack_top,
            dst_kstack_top
        );
        let src_trap_cx_ptr =
            (src_kstack_top - core::mem::size_of::<TrapContext>()) as *const TrapContext;
        let dst_trap_cx_ptr =
            (dst_kstack_top - core::mem::size_of::<TrapContext>()) as *mut TrapContext;
        let task_cx_ptr =
            (dst_trap_cx_ptr as usize - core::mem::size_of::<TaskContext>()) as *mut TaskContext;

        // 复制fd_table和cwd
        let mut child_fd_table: Vec<Option<Arc<dyn File + Send + Sync>>> = Vec::new();
        for fd in parent_inner.fd_table.iter() {
            if let Some(file) = fd {
                child_fd_table.push(Some(file.clone()));
            } else {
                child_fd_table.push(None);
            }
        }
        let child_cwd = parent_inner.cwd.clone();
        // 分配新的tid
        let tid = tid_alloc();

        let task = Arc::new(Task {
            // 对于fork出的子进程, 内核栈是新分配的, 其中只有`TrapContext`是父进程的副本, 然后是初始化的`TaskContext`
            kstack: KernelStack(task_cx_ptr as usize),
            tid,
            inner: SpinNoIrqLock::new(TaskInner {
                memory_set,
                task_status: TaskStatus::Ready,
                parent: Some(Arc::downgrade(self)),
                children: Vec::new(),
                exit_code: 0,
                fd_table: child_fd_table,
                cwd: child_cwd,
            }),
        });

        // 设置子进程的`TrapContext`
        // 设置子进程的`TaskContext`
        let child_tp = Arc::as_ptr(&task) as usize;
        let child_satp = task.inner.lock().memory_set.page_table.token();
        let child_task_cx = context::TaskContext::app_init_task_context(child_tp, child_satp);
        unsafe {
            dst_trap_cx_ptr.write(src_trap_cx_ptr.read());
            task_cx_ptr.write(child_task_cx);
            (*dst_trap_cx_ptr).x[10] = 0;
            (*dst_trap_cx_ptr).x[4] = child_tp;
            log::error!("parent user_sp: {:#x}", (*src_trap_cx_ptr).x[2]);
            log::error!("child user_sp: {:#x}", (*dst_trap_cx_ptr).x[2]);
        }
        log::error!(
            "[Task::fork]task {} tp: {:#x}, satp: {:#x}, kernel_stack: {:#x}",
            tid,
            child_tp,
            child_satp,
            task.kstack.0
        );

        parent_inner.children.push(task.clone());
        task
    }
```

### 2.3 进程切换

在操作系统中, 当一个新进程被创建时, 其初始状态是就绪态(Ready), 并被加入到就绪队列中, 等待 CPU 的调度执行。新进程并不会立刻开始运行, 而是需要由操作系统的调度算法决定何时获得 CPU 的执行权。只有当上一个进程执行完成、时间片耗尽或 CPU 空闲时, 调度程序会根据预设的调度算法(详见 2.4)从就绪队列中选择下一个进程执行。

一旦某个进程被选中并开始在 CPU 上运行, 其状态会从就绪态切换为运行态(Running), 并且此时该进程获得了 CPU 的控制权。操作系统利用定时器中断来分配 CPU 时间片, 以确保多任务并发运行。在这段代码中, 时钟中断的时间间隔为 10 毫秒, 即每个时间片为 10 毫秒。当一个时间片耗尽时, 如果当前运行的进程未完成任务, 时钟中断会被触发。操作系统会暂停该进程的执行, 将其状态更改为就绪态, 并将其重新放回就绪队列, 等待下一次调度。

此外, 如果进程在运行过程中遇到错误或完成其所有任务, 操作系统会将该进程的状态设置为终止态(Zombie)。在终止态下, 进程将释放所占用的资源。
![alx_text](./img/任务切换.png)

### 2.4 进程调度

在调度算法的实现上, RROS 当前采用的是 FIFO(First-In, First-Out)调度算法。作为一种简单直观的进程调度策略, FIFO 的核心思想是按照进程到达调度队列的顺序进行执行。具体而言, 先到达的进程会优先获得 CPU 调度, 后到达的进程则需要等待前面的进程完成后才能开始执行。

FIFO 调度通过维护一个先进先出的队列来管理进程, 每当有新进程到达时, 便将其插入队列尾部。调度器总是从队列头部取出进程并将其分配给 CPU 运行。这种顺序性保证了所有进程都能依次获得执行机会。

## 3.内存管理
### 3.1 地址分配
RROS 地址空间的设计如下图所示：
![alt text](./img/内存空间分布.png)

在 QEMU 平台上, 内核的入口地址被设置在 0x8020_0000。在 xv6 和 rCore-Tutorial 中, 地址空间的划分策略是将 0x8020_0000 以上的区域采用直接映射方式用于内核地址空间, 用户空间与内核空间处于两张页表, 这种设计导致用户程序只要进入内核, 即使不切换到其他任务, 也需要切换页表, 刷新TLB, 这对于性能的损失影响巨大, 为了解决这一限制, 充分利用了页表的灵活性, 将内核地址空间通过偏移映射的方式映射到 RISC-V SV39 规范定义的高位虚拟地址空间, 即 0xFFFF_FFC0_0000_0000 至 0xFFFF_FFFF_FFFF_FFFF。这种映射方式确保每个虚拟地址对应的物理地址都具有固定的偏移量。同时, 用户地址空间被调整为从 0x0 到 0x3F_FFFF_FFFF 的范围, 这样用户在进入内核时就不需要切换页表, 同时又有硬件做检查保证用户不能访问内核空间。

## 4.文件系统
### 4.1 Fat32 文件系统

FAT32(File Allocation Table 32)是一种文件系统格式, 全称为 32 位文件分配表。它是 FAT 文件系统家族的一员, 由微软在 1996 年发布, 用于替代 FAT16。FAT32 因其广泛的兼容性和简单性, 广泛应用于各种存储设备, 如硬盘、U 盘、SD 卡等。

FAT32 的主要特点包括  
‧ 稳定性和兼容性:FAT32 提供了良好的稳定性和兼容性, 能够兼容 Win 9X 及以前版
本的 Windows 操作系统。  
‧ 簇大小：使用比 FAT16 更小的簇(数据存储单元), 从而提高了大容量硬盘上的空间
利用率。  
‧ 分区容量：支持的每个分区容量最大可达到 128TB, 远大于 FAT16 的限制。  
‧ 文件大小限制：单个文件最大支持 4GB, 这对于处理大型文件来说是一个限制。 

FAT32 文件系统的结构主要包括三个部分：  
‧ 引导区：包含文件系统的具体信息, 如 FAT 表个数、每个 FAT 表的大小、每扇区内
的字节数目等。  
‧ 文件分配表区：管理磁盘空间和文件, 保存逻辑盘数据区各簇使用情况信息。  
‧ 数据区：存放用户数据, 以簇为分配单位来使用。  
文件系统在RROS中的抽象: ![alx_text](./img/文件系统架构.png)

### 4.2 核心数据结构
### 4.2.1 FAT32FileSystem 与 FAT32Meta
在 RROS 中, FAT32FileSystem 是 FAT32 文件系统的一个具体实现, 用于表示和管理一个 FAT32 格式的文件系统实例。它通过封装块设备(block_device)和文件系统的根节点(root_inode), 提供了对 FAT32 文件系统的抽象表示和操作接口。
```
pub struct FAT32FileSystem {
    pub block_device: Arc<dyn BlockDevice>,
    pub root_inode: Arc<dyn Inode>,
}
```
FAT32FileSystem 在创建时, 会加载 FAT32 文件系统的元数据, 如引导区信息、FAT 表、根目录等。它需要解析底层块设备中的 FAT 表, 构建文件系统的结构, 并将根目录节点作为 root_inode 进行管理。

FAT32Meta 是用于描述 FAT32 文件系统元信息的一个结构体, 封装了与文件系统布局相关的所有关键参数。这些参数主要由 FAT32 的引导扇区和文件系统信息扇区(FS Info Sector)中提取, 用于初始化和管理文件系统的操作
```
pub struct FAT32Meta {
    pub sector_per_cluster: usize,
    pub fat_count: usize,        // count of FAT
    pub fat_sector_count: usize, // sector count of ONE FAT
    pub fat_start_sector: usize,
    pub data_start_sector: usize, // start sector of data region
    pub total_sector_count: usize,
    pub total_cluster_count: usize, // 整个FAT32文件系统中可用的cluster数目(数据区)

    pub root_cluster_id: usize,
    pub fs_info_sector_id: usize,
    pub backup_sector_id: usize,
}
```
FAT32Meta 结构体各个字段的含义如下：

| 字段名              |               含义                |
| :------------------ | :-------------------------------: |
| sector_per_cluster  |       每个簇包含的扇区数目        |
| fat_count           |            FAT 表数量             |
| fat_sector_count    |      每张 FAT 表占用的扇区数      |
| fat_start_sector    |       FAT 区域的起始扇区号        |
| data_start_sector   |       数据区域的起始扇区号        |
| total_sector_count  |        文件系统的总扇区数         |
| total_cluster_count | 数据区域中可用的簇(Cluster)总数 |
| root_cluster_id     |         根目录所在的簇 ID         |
| fs_info_sector_id   |      文件系统信息扇区的编号       |
| backup_sector_id    |      引导扇区的备份扇区编号       |

### 4.2.2 inode
Inode 是一个抽象的文件系统接口, 用于描述文件系统中一个文件或目录节点(inode)的基本行为。它通过定义一组标准操作, 为具体的文件系统实现提供统一的访问接口。在这个设计中, Inode 是一个线程安全(Send + Sync)的特性对象, 可以通过引用计数(Arc)共享访问。
```
pub trait Inode: Send + Sync {
    fn read<'a>(&'a self, _offset: usize, _buf: &'a mut [u8]) -> usize;
    fn write<'a>(&'a self, _offset: usize, _buf: &'a [u8]) -> usize;
    fn mknod(&self, this: Arc<dyn Inode>, name: &str, mode: InodeMode)
        -> SysResult<Arc<dyn Inode>>;
    fn find(&self, this: Arc<dyn Inode>, name: &str) -> SysResult<Arc<dyn Inode>>;
    fn list(&self, this: Arc<dyn Inode>) -> SysResult<Vec<Arc<dyn Inode>>>;
    fn get_meta(&self) -> Arc<InodeMeta>;
    fn load_children_from_disk(&self, this: Arc<dyn Inode>);
    /// clear the file content, inode still exists
    fn clear(&self);
}
```
索引节点对象由 InodeMeta 结构体表示, InodeMeta 是文件系统中 inode 的核心元数据结构, 涵盖了标识、路径和动态信息管理。通过静态字段和动态字段的分离设计, 它既能高效管理静态信息, 又能灵活处理动态信息, 适应多线程文件系统的需求。它在文件和目录的创建、查询、更新和删除中起到了至关重要的作用, 是文件系统实现的关键组成部分。
```
pub struct InodeMeta {
    pub ino: usize,
    pub mode: InodeMode,
    pub name: String,
    pub path: Path,
    pub inner: FSMutex<InodeMetaInner>,
}
```
InodeMeta 结构体各个字段的含义如下：
| 字段名 |                                           含义                                           |
| :----- | :--------------------------------------------------------------------------------------: |
| ino    |                                    表示 inode 的编号                                     |
| mode   |                             表示文件类型(如普通文件、目录)                             |
| name   |                                        表示文件名                                        |
| path   |                                      表示文件的路径                                      |
| inner  | 包含与 inode 关联的可变元数据(如文件大小、时间戳等), 并通过 FSMutex 进行线程安全保护。 |

### 4.2.3 Dentry

在文件系统中, 目录项是用来管理文件在目录树中信息的核心结构体。它是对文件路径的一种抽象, 通过保存文件的基本信息(如文件名、文件类型、属性、文件大小等), 为文件和目录的组织与访问提供支持。  
文件系统的结构通常是一个以根目录为起点的目录树。目录树的根节点对应文件系统的挂载点(根目录 /), 其子节点依次表示子目录或文件。每个节点在逻辑上都与一个目录项对应, 而每个目录项都指向一个文件或目录的索引节点。索引节点保存了文件或目录的元数据(如位置、权限、大小等)。

在 RROS 文件系统中, 与目录项相关的核心结构体是 FAT32DentryContent 和 FAT32DirEntry。其中这两个结构体分别从不同层面描述了目录项及其内容:    

FAT32DentryContent 用于描述目录项的存储位置及上下文信息, 它关联目录项所在的目录文件并记录目录项在文件中的偏移量。
```
pub struct FAT32DentryContent<'a> {
    // 表示条目所在的文件
    file: &'a mut FAT32File,
    offset: usize,
}
```
FAT32DentryContent 结构体各个字段的含义如下：
| 字段名 |               含义               |
| :----- | :------------------------------: |
| file   |     表示目录项所属的目录文件     |
| offset | 表示当前目录项在 file 中的偏移量 |

FAT32DirEntry 是目录项的核心数据结构, 用于描述单个文件或目录的具体属性。它记录了文件的基本信息, 如文件名、大小、属性、时间戳等。
```
pub struct FAT32DirEntry {
    pub lname: [u16; LNAME_MAXLEN],        // 长文件名, UTF-16 编码
    pub sname: [u8; SNAME_LEN],            // 短文件名, 8.3 格式
    pub attr: u8,                          // 文件属性
    pub crt_time: FAT32Timestamp,          // 创建时间
    pub wrt_time: FAT32Timestamp,          // 修改时间
    pub acc_time: FAT32Timestamp,          // 访问时间
    pub fstcluster: u32,                   // 起始簇号
    pub filesize: u32,                     // 文件大小
}
```

### 4.2.4 FAT32File

FAT32File 是一个用于表示 FAT32 文件系统中文件的核心结构体。它封装了文件的簇分配信息、实际存储的数据簇列表以及文件大小等关键属性, 为文件的读写、操作及管理提供基础。
```
pub struct FAT32File {
    pub fat: Arc<FAT32FileAllocTable>,
    clusters: Vec<usize>,
    size: Option<usize>,
}
```
FAT32File 结构体各个字段的含义如下：
| 字段名   |                                     含义                                     |
| :------- | :--------------------------------------------------------------------------: |
| fat      | 表示与该文件相关的 FAT 表,其中FAT 表记录了文件的簇分配信息, 即文件的存储位置 |
| clusters |                         存储文件内容所在的数据簇链表                         |
| size     |                               表示文件的总大小                               |

### 4.2.5 Path

在 4.2.2 的 inode 模块中已经涉及了 Path 结构体, 这是文件系统中用于表示文件或目录路径的核心数据结构。Path 封装了路径的基本组成部分, 包括路径片段和路径的相对性, 提供了对路径的一种抽象表示。
该结构体不仅对路径进行了封装, 还提供了一系列操作接口, 用于路径的解析、操作和管理。例如, 路径的解析可以通过将字符串形式的路径转换为结构化的路径对象来实现, 进而支持文件系统中复杂的路径处理逻辑。此外, Path 结构体还能区分绝对路径和相对路径, 从而在路径操作时适配不同的上下文。
```
pub struct Path {
    inner: Vec<String>,
    is_relative: bool,
}
```
FAT32File 结构体各个字段的含义如下：
| 字段名      |                       含义                       |
| :---------- | :----------------------------------------------: |
| inner       | 动态数组, 存储路径的每个片段, 表示路径的层级结构 |
| is_relative |              指示路径是否为相对路径              |

## 5.总结与展望

### 5.1 工作总结
1. 实现进程管理和内存管理。
2. 支持Fat32文件系统。
3. 支持多核调度运行。
4. 实现初赛要求的所有系统调用。
5. 实现不同等级的调试日志, 实现 backtrace 机制, 在内核崩溃时能够打印函数调用栈。

### 5.2 经验总结
1. 定义清晰的模块接口, 避免跨模块的直接依赖
2. 在内核中设计轻量级的日志系统, 用于记录系统调用、异常处理和关键事件的执行路径。
   
### 5.3 未来计划
1. 完善进程的调度算法
2. 完善进程间通信机制
3. 将RROS的调度单位扩展为线程